#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#
"""Implementation of several domains.

Domains are used by AVMs to describe groups of values of an attribute for
which a program behaves in the same way.

"""
from collections import deque
from dataclasses import dataclass
from itertools import product
from typing import (
    Optional,
    Generic,
    Tuple,
    AbstractSet,
    List,
    Iterable,
    Union,
    TYPE_CHECKING,
)

from xotl.tools.objects import setdefaultattr, lazy

from .types import Domain, TEq, TOrd
from .ranges import Range


@dataclass(init=False, unsafe_hash=True)
class EquivalenceSet(Generic[TEq]):
    "A simple wrapper over a Python set for items with EqTypeClass"
    values: AbstractSet[TEq]
    __slots__ = ("values", "_sample")

    def __init__(self, values: AbstractSet[TEq]) -> None:
        self.values = frozenset(values)

    def to_interval_set(self) -> "IntervalSet[TOrd]":
        """Convert to a semantically-equivalent interval set.

        This should only be applied for equivalence sets containing orderable
        (Ord) items.

        """
        ranges: List[Range[TOrd]] = []
        values = set(self.values)
        while values:
            val: TOrd = values.pop()  # type: ignore
            ranges.append(Range.new_closed(val, val))
        return IntervalSet(ranges)

    def __bool__(self):
        return bool(self.values)

    def __contains__(self, val: TEq) -> bool:
        return val in self.values

    def __le__(self, other: "EquivalenceSet") -> bool:
        if isinstance(other, IntervalSet):
            return self.to_interval_set() <= other
        if not isinstance(other, EquivalenceSet):
            return NotImplemented
        return self.values <= other.values

    def __lt__(self, other: "EquivalenceSet") -> bool:  # pragma: no cover
        if isinstance(other, IntervalSet):
            return self.to_interval_set() < other
        if not isinstance(other, EquivalenceSet):
            return NotImplemented
        return self.values < other.values

    def __ge__(self, other: "EquivalenceSet") -> bool:  # pragma: no cover
        if isinstance(other, IntervalSet):
            return self.to_interval_set() >= other
        if not isinstance(other, EquivalenceSet):
            return NotImplemented
        return self.values >= other.values

    def __gt__(self, other: "EquivalenceSet") -> bool:  # pragma: no cover
        if isinstance(other, IntervalSet):
            return self.to_interval_set() > other
        if not isinstance(other, EquivalenceSet):
            return NotImplemented
        return self.values > other.values

    def __and__(
        self, other: "EquivalenceSet"
    ) -> "EquivalenceSet":  # pragma: no cover
        if isinstance(other, IntervalSet):
            return self.to_interval_set() & other
        if not isinstance(other, EquivalenceSet):
            return NotImplemented
        return EquivalenceSet(self.values & other.values)

    def diff(
        self, other: "EquivalenceSet"
    ) -> "EquivalenceSet":  # pragma: no cover
        if isinstance(other, IntervalSet):
            return self.to_interval_set().diff(other)
        if not isinstance(other, EquivalenceSet):
            return NotImplemented
        return EquivalenceSet(self.values - other.values)

    __sub__ = difference = diff

    def __or__(
        self, other: "EquivalenceSet"
    ) -> "EquivalenceSet":  # pragma: no cover
        if isinstance(other, IntervalSet):
            return self.to_interval_set() | other
        if not isinstance(other, EquivalenceSet):
            return NotImplemented
        return EquivalenceSet(self.values | other.values)

    def union(
        self, *others: "EquivalenceSet"
    ) -> "EquivalenceSet":  # pragma: no cover
        from operator import or_
        from functools import reduce

        sets: Tuple[EquivalenceSet, ...] = tuple(
            s for s in (self,) + others if s
        )
        values: AbstractSet = reduce(or_, (s.values for s in sets), frozenset())
        return EquivalenceSet(values)

    join = union

    @property
    def sample(self) -> Optional[TEq]:
        if self.values:
            # We cache the result to return the same sample all the time.
            # set.pop() return an arbitrary element.
            return setdefaultattr(
                self, "_sample", lazy(lambda: set(self.values).pop())
            )
        else:
            return None


class DomainMismatch(TypeError):
    pass


@dataclass(init=False, unsafe_hash=True)
class IntervalSet(Generic[TOrd]):
    ranges: Tuple[Range[TOrd], ...]
    __slots__ = ("ranges",)

    def __init__(self, bases: Iterable[Range[TOrd]]) -> None:
        if bases:
            # Normalize the bases, we must keep our individual bases
            # disjoint.
            base, *others = self._extract_all_ranges(bases)
            # Ensure the first base range is not empty
            while not base and others:
                base, *others = others
            if base:
                normalized = base.join(*others)
                assert all(r for r in normalized)
                # `join` will return non-empty bases since `base` is non-empty
                self.ranges = normalized
            else:
                assert not others
                self.ranges = ()
        else:
            self.ranges = ()

    @classmethod
    def _from_ordered_disjoint(
        cls, bases: Iterable[Range[TOrd]]
    ) -> "IntervalSet[TOrd]":
        result = cls([])
        result.ranges = tuple(bases)
        return result

    @classmethod
    def _extract_ranges(
        cls, other: Union["IntervalSet[TOrd]", Range[TOrd]]
    ) -> Tuple[Range[TOrd], ...]:
        if isinstance(other, EquivalenceSet):
            return other.to_interval_set().ranges
        if isinstance(other, IntervalSet):
            return other.ranges
        else:
            assert isinstance(other, Range)
            return (other,)

    @classmethod
    def _extract_all_ranges(
        cls, *args: Iterable[Union["IntervalSet[TOrd]", Range[TOrd]]]
    ) -> Tuple[Range[TOrd], ...]:
        return tuple(
            base
            for arg in args
            for bases in arg
            for base in cls._extract_ranges(bases)
        )

    def __repr__(self):
        return f"IntervalSet({self.ranges!r})"

    @property
    def sample(self) -> Optional[TOrd]:
        if self.ranges:
            return next(
                (r.sample for r in self.ranges if r.sample is not None), None
            )
        else:
            return None

    def __contains__(self, value: TOrd) -> bool:
        return any(value in base for base in self.ranges)

    def __bool__(self):
        return bool(self.ranges)

    def __le__(self, other: "IntervalSet[TOrd]") -> bool:
        if isinstance(other, EquivalenceSet):
            other = other.to_interval_set()
        if not isinstance(other, IntervalSet):
            return NotImplemented
        if not self:
            return True
        if not other:
            return False
        return all(
            any(mine <= their for their in other.ranges) for mine in self.ranges
        )

    def __lt__(self, other: "IntervalSet[TOrd]") -> bool:  # pragma: no cover
        if isinstance(other, EquivalenceSet):
            other = other.to_interval_set()
        if not isinstance(other, IntervalSet):
            return NotImplemented
        return self <= other and not (other <= self)

    def __ge__(self, other: "IntervalSet[TOrd]") -> bool:  # pragma: no cover
        if isinstance(other, EquivalenceSet):
            other = other.to_interval_set()
        if not isinstance(other, IntervalSet):
            return NotImplemented
        return other <= self

    def __gt__(self, other: "IntervalSet[TOrd]") -> bool:  # pragma: no cover
        if isinstance(other, EquivalenceSet):
            other = other.to_interval_set()
        if not isinstance(other, IntervalSet):
            return NotImplemented
        return self >= other and not (other >= self)

    def __eq__(self, other) -> bool:
        if isinstance(other, EquivalenceSet):
            other = other.to_interval_set()
        if not isinstance(other, IntervalSet):
            return NotImplemented
        return self <= other <= self

    def __and__(self, other: "IntervalSet[TOrd]") -> "IntervalSet[TOrd]":
        if isinstance(other, EquivalenceSet):
            other = other.to_interval_set()
        if not isinstance(other, IntervalSet):
            return NotImplemented
        bases = tuple(
            filter(
                None,
                (
                    base & other_base
                    for base, other_base in product(self.ranges, other.ranges)
                ),
            )
        )
        return IntervalSet(bases)

    def union(self, *others: "IntervalSet[TOrd]") -> "IntervalSet[TOrd]":
        all_ranges = self._extract_all_ranges(self.ranges, others)
        if all_ranges:
            base, *bases = all_ranges
            new_bases = base.join(*bases)
        else:
            new_bases = ()
        return IntervalSet(new_bases)

    __or__ = join = union

    def diff(self, other: "IntervalSet[TOrd]") -> "IntervalSet[TOrd]":
        # The algorithm depends that 'ranges' are sorted ranges from left to
        # right by the lowerbound, like this:
        #
        #    ours      |------|   |------|   |--| |-----| |---|
        #  theirs   |---------------|  |-|          |-------|
        #
        # Our __init__ ensures this.
        if isinstance(other, EquivalenceSet):
            other = other.to_interval_set()
        ours = deque(self.ranges)
        theirs = deque(other.ranges)
        results = []
        while ours and theirs:
            top: Range = ours.popleft()
            bottom: Range = theirs.popleft()
            assert top and bottom
            if top == bottom:
                # Nothing to produce here, both sub-ranges are consumed.
                pass
            elif top & bottom:
                # There are four possible cases here (five if we count top ==
                # bottom, but we ruled that out):
                #
                # case:       a     !     b    !    c   !    d
                # ~~~~~~~~~~~~~~~~~~+~~~~~~~~~~+~~~~~~~~+~~~~~~~~~
                # top      |------| !   |----| ! |---|  !   |--|
                # bottom     |--|   ! |---|    !   |---|! |------|
                #
                # The case 'a' will produce a range to "left" that cannot
                # intercept with any other range; so we can put it the
                # results.  The second range may intercept with another of
                # theirs so we put it in the front of ours.  But this case,
                # has two sub-cases:
                #
                #                a.1               a.2
                #   top        |-----|   and    |-----|
                #   bottom     |---|               |--|
                #
                # In the first sub-case we cannot push the diff to the result.
                # In the second we have nothing to push to ours.  But those
                # can be treated the same as the cases 'b' and 'c'.
                diff = top.diff(bottom)
                if diff:
                    if (
                        len(diff) > 1
                    ):  # ->                                   case 'a'.
                        left, right = diff
                        results.append(left)
                        ours.appendleft(right)
                    else:
                        assert len(diff) == 1
                        if (
                            bottom.lowerbound <= top.lowerbound
                        ):  # ->         cases 'b' or a.1
                            ours.appendleft(diff[0])
                        elif (
                            top.upperbound <= bottom.upperbound
                        ):  # ->       cases 'c' or a.2
                            results.append(diff[0])
                            # We need to reintroduce bottom to theirs, because
                            # it can overlap with the next top of 'ours'.
                            theirs.appendleft(bottom)
                        else:
                            assert False
                else:  # ->                                                   case 'd'
                    # We need to reintroduce bottom to theirs, because it can
                    # overlap with the next top of 'ours'.
                    theirs.appendleft(bottom)
            else:
                # There are two cases here:
                #
                #                 'a'               'b'
                #   top             |-----|   and    |-----|
                #   bottom     |---|                        |--|
                #
                if bottom.upperbound <= top.lowerbound:
                    ours.appendleft(top)
                elif top.upperbound <= bottom.lowerbound:
                    theirs.appendleft(bottom)
                else:  # pragma: no cover
                    assert False

        if ours:
            assert not theirs
            results.extend(ours)
        return IntervalSet._from_ordered_disjoint(results)

    __sub__ = difference = diff


if TYPE_CHECKING:  # pragma: no cover
    from datetime import datetime, timedelta

    def check(domain: Domain) -> None:
        return None

    def check_subdomain(domain: Domain[datetime]) -> None:
        return None

    check(EquivalenceSet({1}))
    check(
        Range.new_open_right(
            datetime.now(), datetime.now() + timedelta(1)
        ).lift()
    )
    check_subdomain(
        Range.new_open_right(
            datetime.now(), datetime.now() + timedelta(1)
        ).lift()
    )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#
import pytest

from hypothesis import given, example, strategies as st

from xotless.itertools import CyclicInteger, Line, CyclicLine


@given(st.integers(), st.integers(min_value=1), st.integers())
def test_cyclic_integer_range(x, mod, y):
    assert 0 <= CyclicInteger(x, mod) < mod
    assert 0 <= CyclicInteger(x, mod) + y < mod
    assert 0 <= CyclicInteger(x, mod) - y < mod


@given(st.lists(st.integers(), min_size=0))
@example([])
def test_possible_empty_line(instances):
    l = Line.from_iterable(instances)
    assert list(instances) == list(l)
    l.seek(object())


@given(st.lists(st.integers(), min_size=1))
def test_line(instances):
    l = Line(*instances)
    assert list(instances) == list(l)

    l = Line(*instances)
    assert l.current == instances[0]
    assert l.beginning

    l = Line(*instances, start=-1)
    assert l.ending


@given(st.lists(st.integers(), min_size=2), st.integers())
def test_line_next(instances, pos):
    l = Line(*instances, start=pos)
    pos = l.pos
    try:
        l.next()
    except StopIteration:
        pass
    else:
        assert l.pos == pos + 1


@given(st.lists(st.integers(), min_size=2, max_size=100), st.integers())
def test_line_back(instances, pos):
    l = Line(*instances, start=pos)
    pos = l.pos
    try:
        l.back()
    except StopIteration:
        pass
    else:
        assert l.pos == pos - 1


@given(st.lists(st.integers(), min_size=2, max_size=100), st.integers())
def test_line_back_inverse_of_next(instances, pos):
    l = Line(*instances, start=pos)
    r = ~l
    try:
        previous = l.back()
        try:
            assert r.next() == previous
        except StopIteration:
            assert (
                False
            ), "If I can go back in l, I must be able to forward in ~l"
    except StopIteration:
        pass


@given(st.lists(st.integers(), min_size=2, max_size=100), st.integers())
def test_line_next_inverse_of_back(instances, pos):
    l = Line(*instances, start=pos)
    r = ~l
    try:
        following = l.next()
        try:
            assert r.back() == following
        except StopIteration:
            assert (
                False
            ), "If I can go forward in l, I must be able to back in ~l"
    except StopIteration:
        pass


@given(st.lists(st.integers(), min_size=2, max_size=100))
def test_line_back_after_next(instances):
    l = Line(*instances)
    pos = l.pos
    l.next()  # I know I can do this since instances has at least 2 items
    # After a successful next, we expect we can go back.
    l.back()
    assert l.pos == pos


@given(st.lists(st.integers(), min_size=1, max_size=100))
def test_line_at_end(instances):
    l = Line(*instances, start=-1)
    assert l.ending
    pos = l.pos
    l.next()  # At the end, there's still one item left in the iterator
    assert l.pos == pos + 1
    with pytest.raises(StopIteration):
        l.next()
    assert l.pos == pos + 1, "If next() failed we stay where we were"


@given(st.lists(st.integers(), min_size=1, max_size=100))
def test_line_at_beginning(instances):
    l = Line(*instances, start=0)
    assert l.beginning
    pos = l.pos
    # At the beginning, there's still one item left in the reversed iterator
    l.back()
    assert l.pos == pos - 1
    with pytest.raises(StopIteration):
        l.back()
    assert l.pos == pos - 1, "If back() failed we stay where we were"


@given(st.lists(st.integers(), max_size=100), st.data())
def test_seek1(instances, data):
    if instances:
        positions = st.integers(min_value=0, max_value=len(instances) - 1)
        pos = data.draw(positions)
        pos2 = data.draw(positions)
        what = max(instances) + 1  # this won't be in the line
    else:
        pos = pos2 = 0
        what = -1
    instances[pos:pos] = [what]
    l = Line(*instances, start=pos2)
    l.seek(what)
    assert l.pos == pos


@given(st.lists(st.integers(), max_size=100), st.data())
def test_seek2(instances, data):
    if instances:
        positions = st.integers(min_value=0, max_value=len(instances) - 1)
        pos = data.draw(positions)
        what = max(instances) + 1  # this won't be in the line
    else:
        pos = 0
        what = -1
    l = Line(*instances, start=pos)
    l.seek(what)
    assert l.pos == pos


class _distinct(object):
    pass


@st.composite
def distinct(draw):
    return _distinct()


@given(st.lists(distinct(), min_size=2, max_size=100), st.integers())
def test_line_invert_leave_current(instances, pos):
    l = Line(*instances, start=pos)
    # Notice instances are not integers and that we use `is` instead of `==`,
    # so truly the same object.
    assert l.current is (~l).current


def test_singleton_line():
    l = Line(1)
    assert l
    assert l.beginning
    assert l.ending


def test_empty_line():
    l = Line()
    assert list(l) == []

    l = Line()
    assert list(l) == []

    assert list(~l) == list(l)

    with pytest.raises(IndexError):
        l.current

    assert not l
    assert l.beginning  # beginning is ill-defined for an empty list
    assert not l.ending


@given(st.lists(distinct(), max_size=100), st.integers())
def test_ending_entails_no_posterior(instances, pos):
    l = Line(*instances, start=pos)
    if l.ending:
        with pytest.raises(IndexError):
            l.posterior


@given(st.lists(distinct(), max_size=100), st.integers())
def test_beginning_entails_no_anterior(instances, pos):
    l = Line(*instances, start=pos)
    if l.beginning:
        with pytest.raises(IndexError):
            l.anterior


@given(
    st.lists(distinct(), max_size=50),
    st.lists(distinct(), max_size=50),
    st.integers(),
    st.integers(),
)
def test_line_addition(xs, ys, pos, pos2):
    lxs = Line(*xs, start=pos)
    lys = Line(*ys, start=pos2)
    lxys = lxs + lys
    assert lxs.pos == lxys.pos
    assert list(lxs) + list(lys) == list(lxys)

    lxys = lxs + ys
    assert lxs.pos == lxys.pos
    assert list(lxs) + list(ys) == list(lxys)

    lxys = ys + lxs
    assert lxs.pos == lxys.pos
    assert list(ys) + list(lxs) == list(lxys)


@given(
    st.lists(distinct(), max_size=10),
    st.integers(),
    # the multiplier reasonable small
    st.integers(min_value=-100, max_value=100),
)
def test_line_multiplication(xs, pos, times):
    lxs = Line(*xs, start=pos)
    lxys = lxs * times
    assert lxs.pos == lxys.pos
    assert list(lxs) * times == list(lxys)

    lxys = times * lxs
    assert lxs.pos == lxys.pos
    assert times * list(lxs) == list(lxys)


@given(
    st.lists(st.integers(), min_size=10, max_size=100),
    st.integers(min_value=10, max_value=100),
)
def test_cyclic_line_backandforth_forever(items, times):
    length = len(items)
    l = CyclicLine(*items)
    assert not l.beginning and not l.ending
    for _ in range(times + length):
        l.next()
    l = CyclicLine(*items)
    for _ in range(times + length):
        l.back()


def test_non_possible_empty_cyclic_lines():
    with pytest.raises(TypeError):
        CyclicLine()


@given(st.integers(), st.integers(min_value=1))
def test_cycle_integers_are_pickable(x, l):
    import pickle

    c = CyclicInteger(x, l)
    assert c == pickle.loads(pickle.dumps(c))
    assert c == pickle.loads(pickle.dumps(c, pickle.HIGHEST_PROTOCOL))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#
from enum import Enum
from dataclasses import dataclass
from typing import Optional, Generic, Tuple, List, TypeVar, Type


from xotl.tools.symbols import Undefined
from xotl.tools.infinity import Infinity

from .types import OrdTypeClass, EqTypeClass

TEq = TypeVar("TEq", bound=EqTypeClass)
TOrd = TypeVar("TOrd", bound=OrdTypeClass)


@dataclass(init=False)
class Range(Generic[TOrd]):
    """A range containing values between `lowerbound` and `upperbound`.

    Boundaries can be instances of `Included`:class: or `Excluded`:class:

    Ranges are classified in four `kinds <Range.kind>`:any: according to the
    type of the boundaries.  You can use the four classmethods to build each,
    or you can create the boundaries yourself.

    ===================    =================      ======================
    Lower boundary         Upper boundary         Constructor
    ===================    =================      ======================
    Included               Included               `new_closed`:func:
    Included               Excluded               `new_open_right`:func:
    Excluded               Included               `new_open_left`:func:
    Excluded               Excluded               `new_open`:func:
    ===================    =================      ======================

    In any of the constructors, if `lowerbound` is None, the range's lower
    bound is ``Excluded(-Infinity)``.  If `upperbound` is None, the range's
    upper bound is ``Excluded(Infinity)``.

    Ranges are iterable.  They yield exactly two times: the lower bound and
    the upperbound.  This makes it easy to unpack calling signatures that
    doesn't take ranges but the boundaries.

    Calling `bool` on ranges return True when the range is not empty.

    """

    _lower: "Bound[TOrd]"
    _upper: "Bound[TOrd]"

    __slots__ = ("_lower", "_upper")

    @property
    def lowerbound(self) -> TOrd:
        return self._lower.bound

    @property
    def upperbound(self) -> TOrd:
        return self._upper.bound

    @classmethod
    def new_closed(
        cls, lowerbound: Optional[TOrd], upperbound: Optional[TOrd]
    ) -> "Range[TOrd]":
        """Return a closed range if both boundaries allow it.

        If any of the boundaries is None, or an Infinity value, the result
        will be actually an open or semi-open range.

        """
        lb = Bound.normalized_lower_bound(lowerbound, Included)
        ub = Bound.normalized_upper_bound(upperbound, Included)
        return Range(lb, ub)

    @classmethod
    def new_open(
        cls, lowerbound: Optional[TOrd], upperbound: Optional[TOrd]
    ) -> "Range[TOrd]":
        """Return an open range.

        """
        lb = Bound.normalized_lower_bound(lowerbound, Excluded)
        ub = Bound.normalized_upper_bound(upperbound, Excluded)
        return Range(lb, ub)

    @classmethod
    def new_open_left(
        cls, lowerbound: Optional[TOrd], upperbound: Optional[TOrd]
    ) -> "Range[TOrd]":
        """Return a range open on the lower boundary and possibly closed in the upper boundary.

        """
        lb = Bound.normalized_lower_bound(lowerbound, Excluded)
        ub = Bound.normalized_upper_bound(upperbound, Included)
        return Range(lb, ub)

    @classmethod
    def new_open_right(
        cls, lowerbound: Optional[TOrd], upperbound: Optional[TOrd]
    ) -> "Range[TOrd]":
        """Return a range possibly closed on the lower boundary and open in the upper boundary.

        """
        lb = Bound.normalized_lower_bound(lowerbound, Included)
        ub = Bound.normalized_upper_bound(upperbound, Excluded)
        return Range(lb, ub)

    def __init__(
        self, lowerbound: "Bound[TOrd]", upperbound: "Bound[TOrd]"
    ) -> None:
        lowerbound = (
            lowerbound if lowerbound is not None else Excluded(-Infinity)
        )
        upperbound = (
            upperbound if upperbound is not None else Excluded(Infinity)
        )
        if not isinstance(lowerbound, Bound):
            raise ValueError(
                f"lowerbound ({lowerbound}) is not of type Excluded(), Included() or Unbounded()"
            )
        if not isinstance(upperbound, Bound):
            raise ValueError(
                f"Upperbound ({upperbound}) is not of type Excluded(), Included() or Unbounded()"
            )
        if not lowerbound.bound <= upperbound.bound:
            raise ValueError(f"not {lowerbound} <= {upperbound}")
        if lowerbound.bound == upperbound.bound:
            # Normalize the case of empty range.  All empty ranges will be
            # normalized to Range(Excluded(x), Excluded(x))
            if isinstance(lowerbound, Excluded):
                upperbound = Excluded(upperbound.bound)
            elif isinstance(upperbound, Excluded):
                lowerbound = Excluded(lowerbound.bound)
        self._lower = lowerbound
        self._upper = upperbound

    def __repr__(self):
        return f"Range({self._lower!r}, {self._upper!r})"

    def lift(self):  # pragma: no cover
        "Return the `~xotless.domains.IntervalSet`:class: containing only `self`."
        from .domains import IntervalSet

        return IntervalSet([self])

    def __iter__(self):
        yield self.lowerbound
        yield self.upperbound

    def __bool__(self):
        "True if and only if not empty"
        return not self.empty

    @property
    def empty(self):
        """True if and only if empty"""
        return self.lowerbound == self.upperbound and (
            isinstance(self._lower, Excluded)
            or isinstance(self._upper, Excluded)
        )

    def __contains__(self, value):
        """True if `values` is a member."""
        if self.kind == RangeKind.CLOSED:
            return self.lowerbound <= value <= self.upperbound
        elif self.kind == RangeKind.OPEN:
            return self.lowerbound < value < self.upperbound
        elif self.kind == RangeKind.OPEN_RIGHT:
            return self.lowerbound <= value < self.upperbound
        else:
            assert self.kind == RangeKind.OPEN_LEFT
            return self.lowerbound < value <= self.upperbound

    @property
    def bound(self) -> bool:
        """True if neither boundary is unbounded."""
        return bool(self._lower) and bool(self._upper)

    def diff(self, other: "Range") -> Tuple["Range", ...]:
        """The difference of self from other.

        The result is two ranges that jointly have all the values in self,
        that are not in other.  We need two value of Range because the
        difference is not necessarily a range, but can be fully capture but
        the union of two possibly empty ranges.

        """
        other = self & other
        if self == other:
            return ()
        elif other:
            assert self > other
            res: Tuple[Range[TOrd], ...] = ()
            before = Range(self._lower, ~other._lower)
            if before:
                res += (before,)
            after = Range(~other._upper, self._upper)
            if after:
                res += (after,)
            return res
        else:
            return (self,)

    def _merge_with(self, other: "Range") -> None:
        """Merge an overlapping `other` into self.

        Modify `self` in-place (thus breaking the immutability property) so
        that it becomes the union of `self` and `other`.  It's only valid if
        `self` overlaps with `other`.

        """
        assert self & other or (
            self.upperbound == other.lowerbound
            and (
                isinstance(self._upper, Included)
                or isinstance(other._lower, Included)
            )
        ), f"Cannot merge {self} and {other}"

        # The join of a `Excluded(a)` bound with `Included(a)` bound
        # must allways include a despite the order.
        # So we need find the max and min `bound` and `type`.
        self._lower = min(
            self._lower, other._lower, key=_max_key_by_bound_and_type
        )
        self._upper = max(
            self._upper, other._upper, key=_min_key_by_bound_and_type
        )

    def join(*ranges_: "Range[TOrd]") -> Tuple["Range[TOrd]", ...]:
        """Return several disjoint ranges that describe the union of `ranges`.

        """
        # The join of `Range(Excluded(a), b)` with `Range(Included(a),
        # c)` must include a despite the order.  So we need order by
        # lowerbound and type.
        ranges = sorted((r for r in ranges_ if r), key=_by_lowerbound_and_type)
        result: List[Range] = []
        for next_ in ranges:
            if not result:
                # Don't use the same object since we can modify it with
                # _merge_with.
                result.append(Range(next_._lower, next_._upper))
            else:
                current = result[-1]
                # We have to implement a softer version of __and__ because if
                # r1 = Range(a, b) and r2 = Range(b, c) are disjoint then:
                #
                #              r1 and r2 can be merged to Range(a,c) if:
                #           =============================================
                # r1 = Range(Bound(a), Excluded(b)) or r1 = Range(Bound(a), Included(b))
                # r2 = Range(Included(b), Bound(c)) or r2 = Range(Excluded(b), Bound(c))
                #
                #                  r1 and r2 can't be merged if:
                #         =============================================
                #                   Range(Bound(a), Excluded(b))
                #                   Range(Excluded(b), Bound(c))
                #
                if current & next_ or (
                    current.upperbound == next_.lowerbound
                    and current._upper != next_._lower
                ):
                    current._merge_with(next_)
                else:
                    result.append(Range(next_._lower, next_._upper))
        if result:
            return tuple(result)
        else:
            return ()

    def __and__(self, other: "Range") -> "Range":
        """Get the intersection of self and other."""
        if not isinstance(other, Range):
            return NotImplemented
        lower = max(self._lower, other._lower, key=_max_key_by_bound_and_type)
        upper = min(self._upper, other._upper, key=_min_key_by_bound_and_type)
        if lower.bound <= upper.bound:
            return Range(lower, upper)
        else:
            # empty set in the min value; we need to make Excluded explicit
            return Range(Excluded(upper.bound), Excluded(upper.bound))

    def __eq__(self, other):
        """True if `self` is equal to `other`.

        There are many representations for the empty range, all are equal:

           >>> Range.new_open(0, 0) == Range.new_open(1, 1)
           True

        """
        if not isinstance(other, Range):
            return NotImplemented
        # There are many possible ways to represent the empty range:
        #
        # - Range(Excluded(x), Excluded(x)),
        # - Range(Included(x), Excluded(x)),
        # - Range(Excluded(x), Included(x))
        #
        # for every value of x; but they're all equal.
        if self.empty and other.empty:
            return True
        return self._lower == other._lower and self._upper == other._upper

    def __hash__(self):
        # Since all empty ranges are equal, all Range(x, x) must hash the
        # same.  So, we hash'em all the same as (Undefined, Undefined).
        if self:
            return hash((self._lower, self._upper))
        else:
            return hash((Undefined, Undefined))

    def __le__(self, other: "Range") -> bool:  # pragma: no cover
        "True if self is a sub-range of other"
        try:
            return self == (self & other)
        except TypeError:
            return NotImplemented

    def __lt__(self, other: "Range") -> bool:  # pragma: no cover
        "True is self is a proper sub-range of other"
        try:
            return self != other and self <= other
        except TypeError:
            return NotImplemented

    def __ge__(self, other: "Range") -> bool:  # pragma: no cover
        "True is self is a super-range of other"
        try:
            return (self & other) == other
        except TypeError:
            return NotImplemented

    def __gt__(self, other: "Range") -> bool:  # pragma: no cover
        "True is self is a proper super-range of other"
        try:
            return self != other and self >= other
        except TypeError:
            return NotImplemented

    @property
    def sample(self) -> Optional[TOrd]:
        """Return a member of the range.  Return None if the range is empty or OPEN.

        """
        if self.empty or self.kind == RangeKind.OPEN:
            # There's no general way to get a sample from an OPEN range.
            return None
        if isinstance(self._lower, Included):
            return self.lowerbound
        else:
            assert isinstance(self._upper, Included)
            return self.upperbound

    @property
    def kind(self) -> "RangeKind":
        "The kind of the range."
        if isinstance(self._lower, Included) and isinstance(
            self._upper, Excluded
        ):
            return RangeKind.OPEN_RIGHT
        elif isinstance(self._lower, Included) and isinstance(
            self._upper, Included
        ):
            return RangeKind.CLOSED
        elif isinstance(self._lower, Excluded) and isinstance(
            self._upper, Excluded
        ):
            return RangeKind.OPEN
        else:
            assert isinstance(self._lower, Excluded) and isinstance(
                self._upper, Included
            )
            return RangeKind.OPEN_LEFT

    def __str__(self):
        return f"Range({self._lower!s}, {self._upper!s})"


@dataclass(unsafe_hash=True, order=True)
class Bound(Generic[TOrd]):
    """A boundary condition to be used in Range.

    You must use instances of `Included`:class: or `Excluded`:class:.  Calling
    `bool`:func: on instances of Included always return True.  Calling
    `bool`:func: on instances of Excluded return True when the boundary is
    neither `~xotl.tools.infinity.Infinity`:data: nor -Infinity.

    """

    bound: TOrd

    @classmethod
    def normalized_upper_bound(
        cls, bound: Optional[TOrd], Boundary: Type["Bound[TOrd]"]
    ) -> "Bound[TOrd]":
        """Return a normalized instance of Boundary for the upper bound of Range.

        If `bound` is either None, False, Infinity or -Infinity, return
        ``Excluded(Infinity)``; otherwise return ``Boundary(bound)``.

        """
        assert Boundary is Included or Boundary is Excluded
        assert not isinstance(bound, Bound)
        if bound is None or bound is False or isinstance(bound, type(Infinity)):
            return Excluded(Infinity)
        else:
            return Boundary(bound)

    @classmethod
    def normalized_lower_bound(
        cls, bound: Optional[TOrd], Boundary: Type["Bound[TOrd]"]
    ) -> "Bound[TOrd]":
        """Return a normalized instance of Boundary for the upper bound of Range.

        If `bound` is either None, False, Infinity or -Infinity, return
        ``Excluded(-Infinity)``; otherwise return ``Boundary(bound)``.

        """
        assert Boundary is Included or Boundary is Excluded
        assert not isinstance(bound, Bound)
        if bound is None or bound is False or isinstance(bound, type(Infinity)):
            return Excluded(-Infinity)
        else:
            return Boundary(bound)

    def __invert__(self) -> "Bound[TOrd]":
        ...


class Included(Bound):
    "An included boundary in a `Range`:class:."

    def __init__(self, bound: TOrd) -> None:
        assert bound is not -Infinity and bound is not Infinity
        assert bound is not None and bound is not False
        super().__init__(bound)

    @classmethod
    def safer_new(cls, bound: TOrd) -> "Bound[TOrd]":
        """Try to safely create an Included boundary, falling back to
        `Excluded`:class: for infinite marks.

        """
        if bound is Infinity or bound is -Infinity:
            return Excluded(bound)
        else:
            return cls(bound)

    def __bool__(self):
        return True

    def __repr__(self):
        return f"Included({self.bound!r})"

    def __str__(self):
        return f"={self.bound!s}"

    def __invert__(self):
        return Excluded(self.bound)


class Excluded(Bound):
    "An excluded boundary in a `Range`:class:"

    def __bool__(self):
        if self.bound in (Infinity, -Infinity):
            return False
        else:
            return True

    def __repr__(self):
        return f"Excluded({self.bound!r})"

    def __str__(self):
        return f"≠{self.bound!s}"

    def __invert__(self):
        if self.bound not in (Infinity, -Infinity):
            return Included(self.bound)
        else:
            return self


class RangeKind(Enum):
    """The four kinds of ranges"""

    OPEN_RIGHT = "{lb <= a < ub}"
    OPEN_LEFT = "{lb < a <= ub}"
    CLOSED = "{lb <= a <= ub}"
    OPEN = "{lb < a < ub}"


def _max_key_by_bound_and_type(boundary: Bound[TOrd]) -> Tuple[TOrd, int]:
    """Return a tuple of `(bound, -1 or 1)`.

    The second component is -1 if the boundary is Included, and 1 if Excluded.
    This has the nice property of ordering same-boundary Included *before*
    Excluded.  So:

    >>> max(Included(0), Excluded(0), key=_max_key_by_bound_and_type)
    Excluded(0)

    """
    return (boundary.bound, -1 if isinstance(boundary, Included) else 1)


def _min_key_by_bound_and_type(boundary: Bound[TOrd]) -> Tuple[TOrd, int]:
    """Return a tuple of `(bound, -1 or 1)`.

    This is the same as _max_key_by_bound_and_type but negating the second
    component of the result.

    >>> min(Included(0), Excluded(0), key=_min_key_by_bound_and_type)
    Excluded(0)

    """
    bound, order = _max_key_by_bound_and_type(boundary)
    return bound, -order


def _by_lowerbound_and_type(r: Range[TOrd]) -> Tuple[TOrd, int]:
    return _max_key_by_bound_and_type(r._lower)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#
import pickle
from datetime import datetime
from hypothesis import strategies as st, given

from ..ranges import Range
from ..trees import Cell, IntervalTree
from .strategies.domains import (
    many_date_ranges,
    members_of_date_ranges,
    date_ranges,
)
from .strategies.domains import (
    many_float_ranges,
    members_of_float_ranges,
    float_ranges,
)


MORE_THAN_ONE_LIFETIME = Range.new_open_right(
    datetime(1978, 10, 21), datetime(9078, 10, 21)
)


@given(st.data(), many_date_ranges(5, outer=MORE_THAN_ONE_LIFETIME))
def test_date_interval_trees_finds_the_right_interval(data, intervals):
    cells = [Cell.from_range(r, i) for i, r in enumerate(intervals)]
    tree = IntervalTree.from_cells(cells)
    picked_cell = data.draw(st.sampled_from(cells))
    interval = Range.new_open_right(
        picked_cell.lowerbound, picked_cell.upperbound
    )
    member = data.draw(members_of_date_ranges(interval))
    assert tree[member] == (picked_cell,)
    for protocol in range(pickle.DEFAULT_PROTOCOL, pickle.HIGHEST_PROTOCOL):
        tree = pickle.loads(pickle.dumps(tree, protocol))
        assert tree[member] == (picked_cell,)


@given(st.data(), many_float_ranges(5, outer=Range.new_open_right(-1e10, 1e10)))
def test_float_interval_trees_finds_the_right_interval(data, intervals):
    cells = [Cell.from_range(r, i) for i, r in enumerate(intervals)]
    tree = IntervalTree.from_cells(cells)
    picked_cell = data.draw(st.sampled_from(cells))
    interval = Range.new_open_right(
        picked_cell.lowerbound, picked_cell.upperbound
    )
    member = data.draw(members_of_float_ranges(interval))
    assert tree[member] == (picked_cell,)
    for protocol in range(pickle.DEFAULT_PROTOCOL, pickle.HIGHEST_PROTOCOL):
        tree = pickle.loads(pickle.dumps(tree, protocol))
        assert tree[member] == (picked_cell,)


@given(
    st.data(),
    st.lists(
        float_ranges(
            outer=Range.new_open_right(-1e10, 1e10),
            kinds=st.just(Range.new_open_right),
        ),
        min_size=1,
        max_size=10,
    ),
)
def test_float_interval_trees_finds_only_the_right_intervals(data, intervals):
    cells = [Cell.from_range(r, i) for i, r in enumerate(intervals)]
    tree = IntervalTree.from_cells(cells)
    picked_cell = data.draw(st.sampled_from(cells))
    interval = Range.new_open_right(
        picked_cell.lowerbound, picked_cell.upperbound
    )
    member = data.draw(members_of_float_ranges(interval))
    expected = {cell for cell in cells if member in cell}
    assert set(tree[member]) == expected
    if len(cells):
        assert len(expected)
    for protocol in range(pickle.DEFAULT_PROTOCOL, pickle.HIGHEST_PROTOCOL):
        tree = pickle.loads(pickle.dumps(tree, protocol))
        assert set(tree[member]) == expected


@given(
    st.data(),
    st.lists(
        date_ranges(
            outer=MORE_THAN_ONE_LIFETIME, kinds=st.just(Range.new_open_right)
        ),
        min_size=1,
        max_size=10,
    ),
)
def test_date_interval_trees_finds_only_the_right_intervals(data, intervals):
    cells = [Cell.from_range(r, i) for i, r in enumerate(intervals)]
    tree = IntervalTree.from_cells(cells)
    picked_cell = data.draw(st.sampled_from(cells))
    interval = Range.new_open_right(
        picked_cell.lowerbound, picked_cell.upperbound
    )
    member = data.draw(members_of_date_ranges(interval))
    expected = {cell for cell in cells if member in cell}
    assert set(tree[member]) == expected
    if len(cells):
        assert len(expected)
    for protocol in range(pickle.DEFAULT_PROTOCOL, pickle.HIGHEST_PROTOCOL):
        tree = pickle.loads(pickle.dumps(tree, protocol))
        assert set(tree[member]) == expected



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#
# The idea is to support instances registration; then the inspector can be
# improved to find the right widget for the given instance of a typeclass
# annotation.
from typing import Any  # noqa
from typing import TypeVar, Optional
from typing_extensions import Protocol, runtime


class TypeClass(Protocol):
    pass


# fmt: off
class EqTypeClass(TypeClass, Protocol):  # pragma: no cover
    def __eq__(self, other) -> bool: ...
    def __ne__(self, other) -> bool: ...


class OrdTypeClass(EqTypeClass, Protocol):  # pragma: no cover
    def __le__(self, other) -> bool: ...
    def __lt__(self, other) -> bool: ...
    def __ge__(self, other) -> bool: ...
    def __gt__(self, other) -> bool: ...


# fmt: on


T = TypeVar("T")
S = TypeVar("S", bound="Domain[Any]")
TEq = TypeVar("TEq", bound=EqTypeClass)
TOrd = TypeVar("TOrd", bound=OrdTypeClass)


@runtime
class Domain(Protocol[T]):  # pragma: no cover
    r"""A domain is a generalized type for homogeneous sets.

    It's not iterable because some sets are not iterable (e.g. the open
    interval ``(0-1)``).  There's also no notion of 'the next element'.
    Domains can also represent infinite sets (which are not representable by
    Python's sets).

    All values of a domain are of the same type `T`.  It's an error to mix
    domains of different types of values (``T``) in the operations below.

    """

    @property
    def sample(self: S) -> Optional[T]:
        """Produces a value which is a member of the domain.

        Some non-empty domains cannot produce a sample.  For instance,
        `~xotless.domains.IntervalSet`:class: comprising only open ranges
        cannot produce a sample.  They return None.

        Different calls to `sample` should produce the same value.

        """
        ...

    def union(self: S, *others: S) -> S:
        "Return a domain containing elements of `self` or `other`."
        ...

    def __contains__(self, which: T) -> bool:
        "Return True if `which` is a member."
        ...

    def __sub__(self: S, other: S) -> S:
        "Return a domain containing elements of `self` which are not members of `other`."
        ...

    def __or__(self: S, other: S) -> S:
        "Equivalent to `union`:meth:."
        ...

    def __le__(self: S, other: S) -> bool:
        "Return True if `self` is sub-set of `other`"
        ...

    def __lt__(self: S, other: S) -> bool:
        "Return True if `self` is a proper sub-set of `other`"
        ...

    def __ge__(self: S, other: S) -> bool:
        "Return True if `other` is a sub-set of `self`"
        ...

    def __gt__(self: S, other: S) -> bool:
        "Return True if `other` is a proper sub-set of `self`"
        ...

    def __and__(self: S, other: S) -> S:
        "Return a domain containing common elements of `self` and `other`."
        ...

    def __bool__(self: S) -> bool:
        "Return True if the domain is not empty."
        ...



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#
import doctest
import xotless.walk

from .support import captured_stdout, captured_stderr


def test_doctests():
    with captured_stdout() as stdout, captured_stderr() as stderr:
        failure_count, test_count = doctest.testmod(
            xotless.walk,
            verbose=True,
            optionflags=doctest.NORMALIZE_WHITESPACE | doctest.ELLIPSIS,
            raise_on_error=False,
        )
    if test_count and failure_count:  # pragma: no cover
        print(stdout.getvalue())
        print(stderr.getvalue())
        raise AssertionError("pruned_prefix_walk doctest failed")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#
import collections
from collections import Sequence
from contextlib import contextmanager

from xotl.tools.symbols import Unset

from .pickablenv import EnvironmentData


def tail(iterable, n=1):  # pragma: no cover
    "Return an iterator over the last n items"
    # tail(3, 'ABCDEFG') --> E F G
    return iter(collections.deque(iterable, maxlen=n))


def last(iterable, default=Unset):  # pragma: no cover
    "Return the last item"
    try:
        return next(tail(iterable))
    except StopIteration:
        if default is not Unset:
            return default
        else:
            raise


class Line(Sequence):
    """An immutable list of objects arranged in a line.

    Negative indexes are not wrapped around:

      >>> l = Line(1, 2, 3)
      >>> l[-1]  # doctest: +ELLIPSIS
      Traceback (most recent call last):
      ...
      IndexError: Negative index is not supported

    The line starts at the beginning::

      >>> l.current
      1

    Empty lines are always at the beginning but have no current object::

      >>> l = Line()
      >>> l.beginning
      True

      >>> l.current
      Traceback (...)
      ...
      IndexError: ...

    The constructor allows a single keyword argument `start` so that you may
    place the line at a given position upon construction.  This argument
    supports negative indexes, even though the line itself does not.

    You may also provide the instances in the line as the keyword argument
    `instances` (invalid if any positional argument is provided).

    The inverse of a line is another line with the items in reverse order for
    which the current position gives the same object::

       >>> a = object()
       >>> b = object()
       >>> c = object()

       >>> l = Line(a, b, c)
       >>> l.current is (~l).current is a
       True

       >>> l.next() is a
       True

       >>> l.current is (~l).current is b
       True

    In fact, the function `back()` is defined as if performing a `next()` on
    the inverse and "translating" the change in position back from it.

    """

    START = 0

    def __init__(self, *instances, start=Unset):
        self.instances = instances
        if instances and start is not Unset:
            self.pos = start % len(instances)
        else:
            self.pos = self.START

    @classmethod
    def from_iterable(cls, iterable, start=Unset):
        return cls(*iterable, start=start)

    def __iter__(self):
        # Don't share `self` in iterations.
        return type(self)(*self.instances)

    def __getitem__(self, index):
        if index < 0:  # pragma: no cover
            raise IndexError("Negative index is not supported")
        return self.instances[index]

    def __len__(self):
        return len(self.instances)

    def __bool__(self):
        return len(self) > 0

    __nonzero__ = __bool__

    @property
    def current(self):
        return self[self.pos]

    def __next__(self):
        try:
            result = self.current
            self.pos += 1
            return result
        except IndexError:
            raise StopIteration

    next = __next__

    def __add__(self, other):
        from copy import copy

        result = copy(self)
        result.instances += tuple(other)
        return result

    def __radd__(self, other):
        from copy import copy

        result = copy(self)
        result.instances = tuple(other) + result.instances
        return result

    def __mul__(self, other):
        from copy import copy

        result = copy(self)
        result.instances *= other
        return result

    __rmul__ = __mul__

    def __contains__(self, what):
        return what in self.instances

    def __invert__(self):
        # If we represent a line like: [0, 1, *2, 3] were the * indicates
        # the current position.  Then its invert would be: [3, *2, 1, 0], the
        # position of the invert would that stays at the same element.
        result = type(self)(*reversed(self.instances))
        # We can't use the `start=` keyword argument here because `self.pos`
        # can be outside of the permitted range and since `start=` normalizes
        # it values it would result in a inconsistent state.  If self would be
        # just after the end (or just after the beginning) the inverse just be
        # just before the beginning (or just after the end).
        result.pos = len(self) - self.pos - 1
        return result

    def back(self):
        result = next(~self)
        self.pos -= 1
        return result

    def seek(self, what):
        """Puts the line at the first position where `what` is found.

        If `what` is not found, do nothing.

        """
        try:
            pos = self.instances.index(what)
        except ValueError:
            pass
        else:
            self.pos = pos

    @contextmanager
    def save_excursion(self):
        """A context manager that keeps the 'current' position.

        Example::

        >>> line = Line(1, 2, 3, 4)
        >>> line.seek(1)
        >>> line.current
        1

        >>> with line.save_excursion():
        ...    ahead = line.next()

        >>> with line.save_excursion():
        ...    previous = line.back()
        Traceback (...)
        ...
        StopIteration: ...

        >>> line.current, ahead
        (1, 2)

        """
        pos = self.pos
        try:
            yield self
        finally:
            self.pos = pos

    @property
    def beginning(self):
        """True if the line is at the beginning.

        When beginning is True, there will be no anterior item (IndexError).

        """
        return self.pos == 0

    @property
    def ending(self):
        """True if the line is at the ending.

        When ending is True, there will be no posterior item (IndexError).

        """
        return self.pos == len(self.instances) - 1

    @property
    def posterior(self):
        """The item that is posterior to the current one.

        If there's no posterior item, raise an IndexError.

        """
        with self.save_excursion():
            try:
                self.next()
                return self.next()
            except StopIteration:
                raise IndexError

    @property
    def anterior(self):
        """The item that is anterior to the current one.

        If there's no anterior item, raise an IndexError.

        """
        with self.save_excursion():
            try:
                self.back()
                return self.back()
            except StopIteration:
                raise IndexError


class ModelLine(Line):  # pragma: no cover
    """A line of Odoo models.

    Every item in the line must be an Odoo singleton recordset.

    .. warning:: Pickle support

       Model lines support pickling/unpickling, so you can *safely* put them
       in session-like objects.  However several words of caution are in
       order:

       You must perform all operations (expect pickling/unpickling) within a
       valid xoeuf.odoo.api.Environment manager.

       When you unpickle a model line we defer the actual browsing of Odoo
       recordsets until the first time you access any item.  This is because
       we expect that while loading the session data you may not have entered
       a managed Environment, but while accessing any item you're in your own
       code and, thus, an Environment is probably around.

    """

    EMPTY = {}  # type: ignore

    def __getstate__(self):
        if self.instances:
            envdata = {
                EnvironmentData.from_instance(instance)
                for instance in self.instances
            }
            assert len(envdata) == 1, envdata
            env = tuple(envdata)[0]
            state = {
                "pos": self.pos,
                "instances": [
                    ModelInstance(env, instance._name, instance.id)
                    for instance in self.instances
                ],
            }
            return state
        else:
            return self.EMPTY

    def __getitem__(self, index):
        self._resolve()
        return super(ModelLine, self).__getitem__(index)

    def _resolve(self):
        if len(self):
            result = super(ModelLine, self).__getitem__(0)
            if isinstance(result, ModelInstance):
                # This is the first time we get an item after we got unpickle,
                # we need to convert all to Odoo models now and we need to
                # find the environment.
                env = result.envdata.find()
                self.instances = tuple(
                    env[m._name].browse(m.id) for m in self.instances
                )

    def seek(self, what):
        self._resolve()
        return super(ModelLine, self).seek(what)


class CyclicInteger(int):
    """An integer within the Zn cyclic group.

    Since these integers exist in group with modulus n, you must provide the
    value and the modulus::

      >>> CyclicInteger(1, 10)
      CyclicInteger(1, 10)

    The value is automatically wrapped if needed::

      >>> CyclicInteger(10, 10)
      CyclicInteger(0, 10)

    A cyclic integer is an integer (or maybe a long if you're in Python 2).
    The difference is that when adding a cyclic integer to another number
    it will be confined to the Zn group::

      >>> CyclicInteger(9, 10) + 1
      CyclicInteger(0, 10)

    But this may not be true if the first operand is not a cyclic integer::

      >>> 1 + CyclicInteger(9, 10)
      10...

    """

    def __new__(cls, x, l):
        if l <= 0:
            raise TypeError(
                "Cyclic integers are only defined over strictly positive modulus."
            )
        res = super(CyclicInteger, cls).__new__(cls, x % l)
        res.l = l
        return res

    def __add__(self, y):
        return CyclicInteger(int(self) + y, self.l)

    def __sub__(self, y):
        return CyclicInteger(int(self) - y, self.l)

    def __repr__(self):
        return "CyclicInteger(%s, %s)" % (int(self), self.l)

    def __getnewargs__(self):
        return (int(self), self.l)


class CyclicIntegerDescriptor(object):
    def __init__(self, name):
        self.name = name

    def __get__(self, instance, owner):
        if instance is not None:
            return instance.__dict__[self.name]
        else:
            return self

    def __set__(self, instance, value):
        instance.__dict__[self.name] = CyclicInteger(int(value), len(instance))


class CyclicLine(Line):
    """A line that is circular (or cyclic).

    Cyclic lines must contain at least an element::

       >>> CyclicLine()
       Traceback (...)
       ...
       TypeError: ...

    .. warning:: Calling `next()` will never raise a StopIteration so you
       should never blindly iterate over cyclic lines; they won't stop.

    """

    START = 0
    pos = CyclicIntegerDescriptor("pos")

    # In a cyclic line there's no concept of beginning or ending.  You can
    # always go to the next or previous item.

    @property
    def beginning(self):
        return False

    @property
    def ending(self):
        return False


class CyclicModelLine(CyclicLine, ModelLine):
    pass


class ModelInstance(object):  # pragma: no cover
    # Stands for the model within a ModelLine
    def __init__(self, envdata, model, id):
        self.envdata = envdata
        self._name = model
        self.id = id



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#
import logging

from dataclasses import dataclass
from typing import Tuple


logger = logging.getLogger(__name__)


@dataclass(init=False)
class EnvironmentData:
    """A pickable object to represent Odoo's Environment.

    Odoo's Environment holds a connection to the DB and is not pickable.  This
    object holds only data that is pickable, and when required finds the right
    Environment in the Odoo's local data.

    .. warning:: When using the method `find()` you must be sure there's a
       right environment in Odoo's local data.

    """

    dbname: str
    uid: int
    context: dict  # We expect this to be immutable

    @classmethod
    def from_instance(cls, instance):
        res = getattr(instance, "envdata", None)
        if res is not None:
            return res
        else:
            return cls(instance.env)

    def __init__(self, env):
        if isinstance(env, EnvironmentData):
            self.dbname = env.dbname
        else:
            self.dbname = env.cr.dbname
        self.uid = env.uid
        self.context = dict(env.context)

    def __hash__(self):
        return hash(
            (
                self.dbname,
                self.uid,
                frozenset(
                    self.context
                ),  # Consider only the keys to avoid unhashable values
            )
        )

    def find(self, ignore_user=False, ignore_context=False):
        """Find (in the context) a suitable Odoo environment.

        If both `ignore_user` and `ignore_context` are True any environment
        that matches dbname will be a match.

        In the context of `CURRENT_ENVIRONMENT`:func:, return the given
        environment, but only if its of the same db.

        In the context of `FORCED_ENVIRONMENT`:func:, return the forced
        environment.

        """
        # This is a little dirty trick to avoid creating a new environment for
        # self's db name.  Creating a new Environment requires to open a new
        # cursor to the DB and it won't be the same other modules are using and
        # it'd be leaked.
        from xotl.tools.symbols import Undefined as Last
        from xotl.tools.context import context
        from odoo.api import Environment

        forced = context[_FORCED_ENVIRONMENT].get("env", None)
        if isinstance(forced, Environment):
            return forced

        current = context[_CURRENT_ENVIRONMENT].get("env", None)
        if (
            isinstance(current, Environment)
            and current.cr.dbname == self.dbname
        ):
            return current

        try:
            from odoo.http import request

            env = request.env
            if env and (self.dbname != env.cr.dbname):
                env = None
        except Exception:
            env = None

        if env:
            return env

        envs = iter(getattr(Environment._local, "environments", []))
        env = None
        while not env and env is not Last:
            env = next(envs, Last)
            if env and (
                self.dbname != env.cr.dbname
                or (self.uid != env.uid and not ignore_user)
                or (self.context != dict(env.context) and not ignore_context)
            ):
                env = None
        if env:
            return env
        else:
            args = (self.dbname, self.uid, self.context)
            envs = list(getattr(Environment._local, "environments", []))
            logger.warning(
                "Could not find a proper environment for %r.  Environments: %r",
                args,
                envs,
            )
            raise MissingEnvironment(*args)


@dataclass
class PickableRecordset:
    """A pickable object to represent Odoo's recordsets.

    """

    model: str
    ids: Tuple[int]
    envdata: EnvironmentData

    @classmethod
    def from_recordset(cls, recordset):
        result = cls(
            recordset._name,
            tuple(recordset.ids),
            EnvironmentData.from_instance(recordset),
        )
        result.__dict__["instance"] = recordset  # Memoized already!
        return result

    @property
    def instance(self):
        return self.envdata.find()[self.model].browse(self.ids)

    def find_instance(self, ignore_env_user=False, ignore_env_context=False):
        return self.envdata.find(
            ignore_user=ignore_env_user, ignore_context=ignore_env_context
        )[self.model].browse(self.ids)

    # Recordset are equal regardless of the Environment full data.  I don't
    # want to rely on Odoo's implementation of Model's __eq__ and __hash__
    # because they assume (actually know) inter-database is not possible.  But
    # since this a pickable record we may get an instance of a DB in the
    # context of another.
    def __eq__(self, other):
        if isinstance(other, PickableRecordset):
            return (
                self.model == other.model
                and set(self.ids) == set(other.ids)
                and self.envdata.dbname == other.envdata.dbname
            )
        else:
            return NotImplemented

    def __hash__(self):
        return hash((self.model, frozenset(self.ids), self.envdata.dbname))

    def __getstate__(self):
        return self.model, self.ids, self.envdata

    def __setstate__(self, state):
        self.model, self.ids, self.envdata = state


class MissingEnvironment(RuntimeError):
    pass


_FORCED_ENVIRONMENT = object()


def FORCED_ENVIRONMENT(env):
    """A context manager to force a given environment while unpickling.
    """
    from odoo.api import Environment
    from xotl.tools.context import context

    assert isinstance(env, Environment)
    return context(_FORCED_ENVIRONMENT, env=env)


_CURRENT_ENVIRONMENT = object()


def CURRENT_ENVIRONMENT(env):
    """A context manager to use a given environment while unpickling.
    """
    from odoo.api import Environment
    from xotl.tools.context import context

    assert isinstance(env, Environment)
    return context(_CURRENT_ENVIRONMENT, env=env)
